import asyncio

import aiohttp
import discord
from async_rediscache import RedisSession
from discord.ext import commands
from pydis_core import StartupError
from pydis_core.utils.logging import get_logger
from redis import RedisError

import sir_lancebot as bot  # Change here to import the main package
from sir_lancebot import constants  # Update to import constants from the new package
from sir_lancebot.sir_lancebot import Bot  # Update to import Bot from the new package
from sir_lancebot.log import setup_sentry  # Update to import setup_sentry from the new package
from sir_lancebot.utils.decorators import whitelist_check  # Update to import decorators from the new package

log = get_logger(__name__)
setup_sentry()


async def _create_redis_session() -> RedisSession:
    """Create and connect to a redis session."""
    redis_session = RedisSession(
        host=constants.Redis.host,
        port=constants.Redis.port,
        password=constants.Redis.password.get_secret_value(),
        max_connections=20,
        use_fakeredis=constants.Redis.use_fakeredis,
        global_namespace="bot",
        decode_responses=True,
    )
    try:
        return await redis_session.connect()
    except RedisError as e:
        raise StartupError(e)


async def test_bot_in_ci(bot: Bot) -> None:
    """
    Attempt to import all extensions and then return.

    This is to ensure that all extensions can at least be
    imported and have a setup function within our CI.
    """
    from pydis_core.utils._extensions import walk_extensions

    from sir_lancebot import exts  # Update the import for exts

    for _ in walk_extensions(exts):
        # walk_extensions does all the heavy lifting within the generator.
        pass


async def main() -> None:
    """Entry async method for starting the bot."""
    allowed_roles = list({discord.Object(id_) for id_ in constants.MODERATION_ROLES})
    intents = discord.Intents.default()
    intents.bans = False
    intents.integrations = False
    intents.invites = False
    intents.message_content = True
    intents.typing = False
    intents.webhooks = False

    async with aiohttp.ClientSession() as session:
        bot.instance = Bot(
            guild_id=constants.Client.guild,
            http_session=session,
            redis_session=await _create_redis_session(),
            command_prefix=commands.when_mentioned_or(constants.Client.prefix),
            activity=discord.Game(name=f"Commands: {constants.Client.prefix}help"),
            case_insensitive=True,
            allowed_mentions=discord.AllowedMentions(everyone=False, roles=allowed_roles),
            intents=intents,
            allowed_roles=allowed_roles,
        )

        async with bot.instance as _bot:
            _bot.add_check(whitelist_check(
                channels=constants.WHITELISTED_CHANNELS,
                roles=constants.STAFF_ROLES,
            ))
            if constants.Client.in_ci:
                await test_bot_in_ci(_bot)
            else:
                await _bot.start(constants.Client.token.get_secret_value())


asyncio.run(main())

